==> apracjs-13.1.29.tgz: package/lib/handlers/gspx/parserCache.js <==
﻿
/*!
 * apracjs
 * Copyright(c) 2017 by wssf
 * 2017-06-09
 * 处理gspx文件的转换和缓存
 */

'use strict'
var fs = require('fs');
var debug = require('debug')('aprac:gspxCache');
var htmlparser2 = require('htmlparser2');
var utils = require('../../common/utils.js');
var gConfig = require('../../config.js');
var config = gConfig.apracConfig;
var settings = gConfig.apracSettings;
var sysConsts = require('../../scripts/sysConsts.json');
var PagerInfo = require('../../scripts/dataSource/pagerInfo.js');
var PagerInfoData = PagerInfo.PagerInfoData;
var ConfigStore = require('../../scripts/configStore.js');
var cacheManager = require('../../common/cacheManager.js');

var gspxCaches = {};
exports.isCached = function(filename) {
  return gspxCaches[filename];
}

exports.getGspxCache = function(filename) {
  var cacheItem = gspxCaches[filename];
  if (cacheItem !== undefined) {
    return cacheItem; // 直接取缓存，第二次不需要重复检查并读取文件
  }

  var exists = fs.existsSync(filename); // 第一次读取，必须阻塞等待，避免多个线程重复操作
  if (!exists) {
    debug('当前解析的页面不存在:%s', filename);
    return null;
  }

  cacheItem = new GspxCache(); // 每个页面对应一个缓存容器
  var rootNode;
  var stack = [];
  var node;

  var byte = fs.readFileSync(filename);
  var gspxXML = utils.getUTF8String(byte);

  function nodeExpatToXml(xml) { //通过node-expat解析
    var nodexml = require('../../../modules/node-expat/node-expat.js');
    var parser = new nodexml.Parser('UTF-8');

    parser.on('startElement', function(name, attrs) {
      node = {};
      node.Name = name;
      node.Attributes = attrs;
      stack.push(node);

      cacheItem.parserNode(name, utils.copyTo(attrs, {}));
    });

    parser.on('endElement', function(name) {
      var pNode = stack.pop();
      if (stack.length == 0) {
        rootNode = pNode;
      } else {
        var parentNode = stack[stack.length - 1];
        if (!parentNode.Childs) parentNode.Childs = [];
        parentNode.Childs.push(pNode);
      }
    });

    parser.on('text', function(text) {
      if (text != null && text.trim() != '') {
        node.Value = text;
      }
    });

    parser.on('error', function(error) {
      console.log("read " + filename + "    " + error);
    });
    parser.write(xml);
  }

  function HtmlParser2ToXml(xml) {
    var parser = new htmlparser2.Parser({
      onopentag: function(name, attrs) {
        node = {};
        node.Name = name;
        node.Attributes = attrs;
        stack.push(node);

        cacheItem.parserNode(name, utils.copyTo(attrs, {}));
      },
      ontext: function(text) {
        if (text != null && text.trim() != '') {
          node.Value = text;
        }
      },
      onclosetag: function(tagname) {
        var pNode = stack.pop();
        if (stack.length == 0) {
          rootNode = pNode;
        } else {
          var parentNode = stack[stack.length - 1];
          if (!parentNode.Childs) parentNode.Childs = [];
          parentNode.Childs.push(pNode);
        }
      },
      onerror: function(error) {
        console.log("read " + filename + "    " + error);
      }
    }, {
        lowerCaseTags: false,
        lowerCaseAttributeNames: false,
        xmlMode: true
      });
    parser.write(gspxXML);
    parser.end();
  }

  HtmlParser2ToXml(gspxXML);
  //nodeExpatToXml(gspxXML);
  cacheItem.json = rootNode;
  if (config.env) { // 开发环境要缓存
    gspxCaches[filename] = cacheItem;
  }
  return cacheItem;
}

class GspxCache {
  constructor() {
    this.json = null;
    this.tagCache = [];
    this.hasConfigData = false;
  }

  parserNode(name, attr) { // 解析一些服务端控件：Grid\ListView\FlashChart\Chart\Scheduler,目前主要是Grid
    if (name == "Grid" || name == "TreeGrid") {
      this.tagCache.push(new Grid(attr));
    } else if (name == "ListView") {
      this.tagCache.push(new ListView(attr));
    } else if (name == "FlowPanel") {
      this.tagCache.push(new FlowPanel(attr));
    } else if (name == "Style") {
      this.tagCache.push(new Style(attr));
    } else if (name == "VSplitter" || name == "HSplitter") {
      this.hasConfigData = true; // 优化，存在Splitter才告诉page，需要读取配置，carpa3是每次都要读取配置（即便没有Splitter）
    }
  }

  async initContextSession(pageTag) {
    for (var i = 0, count = this.tagCache.length; i < count; i++) {
      var tag = this.tagCache[i];
      await tag.init(pageTag);
    }

    if (this.hasConfigData) {
      pageTag.addConfigData(pageTag.getInternalID(), await ConfigStore.loadData(
        pageTag.req, pageTag.getInternalID() + "_0", "form"));
    }
  }
}

class BaseStruct {
  constructor(attr) {
    //var regEx = new RegExp('\{(.+?)\}');

    for (var name in attr) {
      var value = attr[name].trim();
      if (!value.startsWith("$")) {
        this[name] = value;
      } else {
        name = "@" + name; // 变量名单独存放格式
        var i1 = value.indexOf('{') + 1;
        var i2 = value.indexOf('}');
        this[name] = value.substr(i1, i2 - i1).trim();
      }
    }
  }

  getValue(name, defaultValue) {
    var value = this[name];
    if (value !== undefined) {
      if (typeof defaultValue == "number") {
        return parseInt(value); // 按类型返回
      } else if (typeof defaultValue == "boolean") {
        return value.toLowerCase() == "true" ? true : false;
      } else {
        return value; // 其它类型直接返回
      }
    }

    value = this.getVariableValue(name);
    if (value === undefined || this._contextData[value] === undefined) {
      return defaultValue == undefined ? null : defaultValue; // 没有变量则返回默认值
    }

    return this._contextData[value]; // 从context获取数据源
  }

  getVariableValue(name) {
    name = "@" + name;
    return this[name]; // 表达式 to do
  }

  init(pageTag) {
    this._page = pageTag;
    this._contextData = pageTag.httpContext.getContextData();
  }

  getID() {
    if (this._id) return this._id;

    this._id = this.makeID(); // 只取一次
    if (!this._id) {
      var name = utils.lowerFirstChar(this._scriptTypeName);
      this._id = this._page.getControlID(name); // 会自动添加 1,2,3 ...
    }

    return this._id;
  }

  getGUID() {
    if (this._guid) return this._guid;
    return this._guid = this._page.getInternalID() + "_" + this.getID(); // 使用InternalID以与地址无关，这样相同页面都使用一样的配置文件
  }

  makeID() {
    var _id = this.ID;
    var _dataField = this.DataField;
    return _id ? _id : (_dataField ? _dataField.Replace('.', '_') : null);
  }
}

class Grid extends BaseStruct {
  constructor(attr) {
    super(attr);

    this._scriptTypeName = "Grid";
    this._itemCountCache = -1;
    this._cDefaultGridHeaderHeight = 20;
    this._cHScrollBarHeight = 17;
    this._cFooterHeight = 21;
    this._defaultRowHeight = 22; //Settings.GridDefaultRowHeight;   // Grid行高默认是21px，由于部分皮肤里面自定义行高，而AutoPagerPageSize按21计算导致结果不正确
  }

  async init(pageTag) {
    super.init(pageTag);

    var saveConfig = super.getValue("SaveConfig", true);
    var allowConfig = super.getValue("AllowConfig", true);
    var hasPager = true;

    var dataSource = super.getValue("DataSource");
    if (!dataSource || super.getValue("Pager", "None") == "None" ||
      typeof dataSource.getItemCount != "function") {
      hasPager = false; // 没有分页器+不是分页数据源
    }

    if (saveConfig && (allowConfig || hasPager)) {
      pageTag.addConfigData(this.getID(), await this.getConfigData()); // 列配置必须要改为同步，不然没法获取自定义的PageSize
    }

    if (!hasPager) {
      return;
    }

    var columnExpands = PagerInfo.getDataSourceColumnExpands(dataSource); // 动态扩展列
    pageTag.addGridColumnExpandDatas(this.getID(), columnExpands);
    let pageClass = pageTag.handler._getPageClass(pageTag.req, {}); // 注入拦截器
    if (pageClass && pageClass.Interceptor) {
      dataSource.Interceptor = new pageClass.Interceptor(pageTag.req);
    }
    var key = super.getVariableValue("DataSource");
    this._contextData[key] = this.initPagersDataSource(pageTag, dataSource); // 第一页数据
  }

  async getConfigData() {
    var saveConfig = super.getValue("SaveConfig", true);
    if (!saveConfig) return null;

    this._configData = await ConfigStore.loadData(this._page.req, this.getGUID() + "_" + super.getValue("ConfigID", 0), "grid");
    return this._configData;
  }

  initPagersDataSource(pageTag, dataSource) {
    var bottomPagerID = this.getID() + "_pager1";
    var topPagerID = this.getID() + "_pager2";
    var pager = this.getValue("Pager", "None");
    if (pager == "TopAndBottom") {
      this.initDataSource(pageTag, topPagerID, dataSource); // 顶部pager初始化（必须同时初始化两个pager）
    } else if (pager == "Top") {
      bottomPagerID = topPagerID;
    }

    return this.initDataSource(pageTag, bottomPagerID, dataSource);
  }

  initDataSource(pageTag, pagerId, dataSource) {
    var pager = new Pager(pageTag);
    var newPageSize = this.getPageSize();
    if (this.getValue("PagerShowAll", false)) {
      newPageSize = -1;
    }
    pager._pageSize = newPageSize;
    pager._id = pagerId;
    pager._dataSource = dataSource;

    pageTag.addGridItemCountDatas(pagerId, pager.getItemCount()); // 数据总行数，翻页控件需要获取（翻页器1）
    return pager.getFirstPageItemList(); // 服务端预先返回第一页数据
  }

  getPageSize() {
    if (this._configData != null && this._configData.PageSize > 0) {
      return this._configData.PageSize; // 用户自定义每页显示条数
    }

    if (this.getValue("AutoPagerPageSize", false) || this.getValue(
      "AutoDefaultRowCount", false)) {
      var height = this.getValue("Height", 300);
      if (height > 0) // 不包括边距的
      {
        var bodyHeight = height - this.getNonBodyHeight() - 1 - this._cHScrollBarHeight; // 1个像素是第一行的顶部线条
        var bodyRows = parseInt((bodyHeight + 1) / this._defaultRowHeight); // 加1像素说明允许超出1像素
        return bodyRows > 1 ? bodyRows : 2;
      }
    }
    return this.getValue("PagerPageSize", settings.GridPagerDefaultPageSize);
  }

  getNonBodyHeight() {
    return this._cDefaultGridHeaderHeight + // 表头
      (this.getValue("ShowHScrollBar", false) ? this._cHScrollBarHeight : 0) + // 加横向滚动条高度
      (this.getValue("ShowFooter", false) ? this._cFooterHeight : 0);
  }
}

class ListView extends BaseStruct {
  constructor(attr) {
    super(attr);
    this._scriptTypeName = "ListView";
  }

  init(pageTag) {
    super.init(pageTag);
  }
}

class FlowPanel extends BaseStruct {
  constructor(attr) {
    super(attr);
    this._scriptTypeName = "FlowPanel";
  }

  async init(pageTag) {
    super.init(pageTag);
    pageTag.addConfigData(this.getID(), await this.getConfigData());
  }

  async getConfigData() {
    var saveConfig = super.getValue("AllowConfig", false);
    if (!saveConfig) return null;

    this._configData = await ConfigStore.loadData(this._page.req, this.getGUID() + "_" + super.getValue("ConfigID", 0), "panel");
    return this._configData;
  }
}

class Style extends BaseStruct {
  constructor(attr) {
    super(attr);
  }

  init(pageTag) {
    super.init(pageTag);
    pageTag.addStyle(this);
  }
}

class Pager {
  constructor(pageTag) {
    this._pageTag = pageTag;
    this._itemCountCache = -1;
    this._showAll = false;
    this._inited = false;
  }

  getItemCount(dataSource) {
    this.initDataSource();
    if (this._itemCountCache < 0) { // 避免在 ShowAll 情况下，在Grid中调用一次，Pager中又调用一次
      this._itemCountCache = this._dataSource.getItemCount();
    }
    return this._itemCountCache;
  }

  getFirstPageItemList() {
    this.initDataSource();
    this._dataSource.invalidateCache(); // 第一次取都是重新刷新
    var newPageSize = this._pageSize;
    if (this._showAll) {
      newPageSize = this.getItemCount();
    }
    var interceptor = this._dataSource && this._dataSource.Interceptor ? this._dataSource.Interceptor : null;
    return this.wrapInterceptor(interceptor, 'getItemList', this._dataSource, [0, newPageSize])
    //return this._dataSource.getItemList(0, newPageSize);
  }

  wrapInterceptor(interceptor, funname, obj, args) {
    return async (cb) => {
      try {
        if (interceptor) {
          await interceptor.before(obj, funname, args);
        }
        let rt = await obj[funname].apply(obj, args);
        if (typeof rt === 'function') {
          await rt(async (result, errormsg) => {
            try {
              if (interceptor) {
                await interceptor.after(obj, funname, args, result, errormsg);
              }
              cb(result, errormsg);
            } catch (e) {
              cb(null, e);
            }
          });
        } else {
          cb(rt);
        }
      } catch (e) {
        cb(null, e);
      }
    }
  }

  initDataSource() {
    if (this._inited) return;
    this._inited = true;

    var page = this._pageTag;
    var httpContext = page.httpContext;
    var req = httpContext.req;
    var session = req.session;
    var dataSource = this._dataSource;
    var clientId = page.getControlClientID(this._id);
    dataSource._pageUrl = req.parsed.path;

    var key = sysConsts.SessionPrefixPagerInfo + clientId;

    var pagerInfoData = req.session[key];
    if (pagerInfoData != null) {
      pagerInfoData._createdCount++;
      pagerInfoData._cacheId = session.id + clientId;
      if (req.headers["server-reqesut"] !== "true") {
        pagerInfoData._oldQueryParams = null;
        pagerInfoData._oldItemCount = 0;
        pagerInfoData._isFirstLoad = true;
      }
      cacheManager.add(pagerInfoData._cacheId, new PagerInfo(session, clientId, dataSource)); // 有时候session还在，但缓存丢失了，没找到重现规律，重新加一次缓存(覆盖的)
    } else {
      pagerInfoData = new PagerInfoData(req.session, clientId, dataSource, req.url);
    }
    req.session[key] = pagerInfoData; // 只能存key-value，不能存function，因此改为：PagerInfoData
  }
}

==> apracjs-18.1.2.tgz: package/lib/handlers/gspx/parserCache.js <==
﻿
/*!
 * apracjs
 * Copyright(c) 2017 by wssf
 * 2017-06-09
 * 处理gspx文件的转换和缓存
 */

'use strict'
var fs = require('fs');
var debug = require('debug')('aprac:gspxCache');
var htmlparser2 = require('htmlparser2');
var utils = require('../../common/utils.js');
var gConfig = require('../../config.js');
var config = gConfig.apracConfig;
var settings = gConfig.apracSettings;
var sysConsts = require('../../scripts/sysConsts.json');
var PagerInfo = require('../../scripts/dataSource/pagerInfo.js');
var PagerInfoData = PagerInfo.PagerInfoData;
var ConfigStore = require('../../scripts/configStore.js');
var cacheManager = require('../../common/cacheManager.js');
var gspxCaches = {};

exports.isCached = function(fileName) {
  return gspxCaches[fileName];
}

exports.getGspxCache = function(fileName) {
  var cacheItem = gspxCaches[fileName];
  if (cacheItem !== undefined) {
    return cacheItem; // 直接取缓存，第二次不需要重复检查并读取文件
  }

  var exists = fs.existsSync(fileName); // 第一次读取，必须阻塞等待，避免多个线程重复操作
  if (!exists) {
    debug('当前解析的页面不存在:%s', fileName);
    return null;
  }

  cacheItem = new GspxCache(); // 每个页面对应一个缓存容器
  var jsonData = HtmlParser2ToJson(fileName, cacheItem);
  cacheItem.jsonstr = JSON.stringify(jsonData);
  if (config.env) { // 正式环境才缓存，开发环境使用物理文件
    gspxCaches[fileName] = cacheItem;
  }
  return cacheItem;
}

function HtmlParser2ToJson(fileName, cacheItem) {
  var stack = [];
  var node = null;
  var rootNode = null;
  var gspxXML = utils.getUTF8String(fs.readFileSync(fileName));

  var parser = new htmlparser2.Parser({
    onopentag: function(name, attrs) {
      node = {};
      node.Name = name;
      node.Attributes = attrs;
      stack.push(node);

      cacheItem.parserNode(name, utils.copyTo(attrs, {}));
    },
    ontext: function(text) {
      if (text != null && text.trim() != '') {
        node.Value = text;
      }
    },
    onclosetag: function(tagname) {
      var pNode = stack.pop();
      if (stack.length == 0) {
        rootNode = pNode;
      } else {
        var parentNode = stack[stack.length - 1];
        if (!parentNode.Childs) parentNode.Childs = [];
        parentNode.Childs.push(pNode);
      }
    },
    onerror: function(error) {
      console.log("read " + fileName + "    " + error);
    }
  }, {
    lowerCaseTags: false,
    lowerCaseAttributeNames: false,
    xmlMode: true
  });
  parser.write(gspxXML);
  parser.end();
  return rootNode;
}


/*
function nodeExpatToXml(xml) { //通过node-expat解析
  var nodexml = require('../../../modules/node-expat/node-expat.js');
  var parser = new nodexml.Parser('UTF-8');

  parser.on('startElement', function(name, attrs) {
    node = {};
    node.Name = name;
    node.Attributes = attrs;
    stack.push(node);

    cacheItem.parserNode(name, utils.copyTo(attrs, {}));
  });

  parser.on('endElement', function(name) {
    var pNode = stack.pop();
    if (stack.length == 0) {
      rootNode = pNode;
    } else {
      var parentNode = stack[stack.length - 1];
      if (!parentNode.Childs) parentNode.Childs = [];
      parentNode.Childs.push(pNode);
    }
  });

  parser.on('text', function(text) {
    if (text != null && text.trim() != '') {
      node.Value = text;
    }
  });

  parser.on('error', function(error) {
    console.log("read " + fileName + "    " + error);
  });
  parser.write(xml);
}
*/


class GspxCache {
  constructor() {
    this.json = null;
    this.tagCache = [];
    this.hasConfigData = false;
  }

  parserNode(name, attr) { // 解析一些服务端控件：Grid\ListView\FlashChart\Chart\Scheduler,目前主要是Grid
    if (name == "Grid" || name == "TreeGrid") {
      this.tagCache.push(new Grid(attr));
    } else if (name == "ListView") {
      this.tagCache.push(new ListView(attr));
    } else if (name == "FlowPanel") {
      this.tagCache.push(new FlowPanel(attr));
    } else if (name == "Style") {
      this.tagCache.push(new Style(attr));
    } else if (name == "VSplitter" || name == "HSplitter") {
      this.hasConfigData = true; // 优化，存在Splitter才告诉page，需要读取配置，craba3是每次都要读取配置（即便没有Splitter）
    }
  }

  async initContextSession(pageTag) {
    for (var i = 0, count = this.tagCache.length; i < count; i++) {
      var tag = this.tagCache[i];
      await tag.init(pageTag);
      tag.dispose();
    }

    if (this.hasConfigData) {
      pageTag.addConfigData(pageTag.getInternalID(), await ConfigStore.loadData(pageTag.req, pageTag.getInternalID() + "_0", "form"));
    }
  }
}

class BaseStruct {
  constructor(attr) {
    //var regEx = new RegExp('\{(.+?)\}');

    for (var name in attr) {
      var value = attr[name].trim();
      if (!value.startsWith("$")) {
        this[name] = value;
      } else {
        name = "@" + name; // 变量名单独存放格式
        var i1 = value.indexOf('{') + 1;
        var i2 = value.indexOf('}');
        this[name] = value.substr(i1, i2 - i1).trim();
      }
    }
  }

  dispose() {
    this._page = null;
    this._contextData = null;
  }

  getValue(name, defaultValue) {
    var value = this[name];
    if (value !== undefined) {
      if (typeof defaultValue == "number") {
        return parseInt(value); // 按类型返回
      } else if (typeof defaultValue == "boolean") {
        return value.toLowerCase() == "true" ? true : false;
      } else {
        return value; // 其它类型直接返回
      }
    }

    value = this.getVariableValue(name);
    if (value === undefined || this._contextData[value] === undefined) {
      return defaultValue == undefined ? null : defaultValue; // 没有变量则返回默认值
    }

    return this._contextData[value]; // 从context获取数据源
  }

  getVariableValue(name) {
    name = "@" + name;
    return this[name]; // 表达式 to do
  }

  init(pageTag) {
    this._page = pageTag;
    this._contextData = pageTag.httpContext.getContextData();
  }

  getID() {
    if (this._id) return this._id;

    this._id = this.makeID(); // 只取一次
    if (!this._id) {
      var name = utils.lowerFirstChar(this._scriptTypeName);
      this._id = this._page.getControlID(name); // 会自动添加 1,2,3 ...
    }

    return this._id;
  }

  getGUID() {
    if (this._guid) return this._guid;
    return this._guid = this._page.getInternalID() + "_" + this.getID(); // 使用InternalID以与地址无关，这样相同页面都使用一样的配置文件
  }

  makeID() {
    var _id = this.ID;
    var _dataField = this.DataField;
    return _id ? _id : (_dataField ? _dataField.Replace('.', '_') : null);
  }
}

class Grid extends BaseStruct {
  constructor(attr) {
    super(attr);

    this._scriptTypeName = "Grid";
    this._itemCountCache = -1;
    this._cDefaultGridHeaderHeight = 20;
    this._cHScrollBarHeight = 17;
    this._cFooterHeight = 21;
    this._defaultRowHeight = 22; //Settings.GridDefaultRowHeight;   // Grid行高默认是21px，由于部分皮肤里面自定义行高，而AutoPagerPageSize按21计算导致结果不正确
  }

  async init(pageTag) {
    super.init(pageTag);

    var saveConfig = super.getValue("SaveConfig", true);
    var allowConfig = super.getValue("AllowConfig", true);
    var hasPager = true;

    var dataSource = super.getValue("DataSource");
    if (!dataSource || super.getValue("Pager", "None") == "None" ||
      typeof dataSource.getItemCount != "function") {
      hasPager = false; // 没有分页器+不是分页数据源
    }

    if (saveConfig && (allowConfig || hasPager)) {
      pageTag.addConfigData(this.getID(), await this.getConfigData()); // 列配置必须要改为同步，不然没法获取自定义的PageSize
    }

    if (!hasPager) {
      return;
    }

    var columnExpands = PagerInfo.getDataSourceColumnExpands(dataSource); // 动态扩展列
    pageTag.addGridColumnExpandDatas(this.getID(), columnExpands);
    let pageClass = pageTag.handler._getPageClass(pageTag.req, {}); // 注入拦截器
    if (pageClass && pageClass.Interceptor) {
      dataSource.Interceptor = new pageClass.Interceptor(pageTag.req);
    }
    var key = super.getVariableValue("DataSource");
    this._contextData[key] = this.initPagersDataSource(pageTag, dataSource); // 第一页数据
  }

  async getConfigData() {
    var saveConfig = super.getValue("SaveConfig", true);
    if (!saveConfig) return null;

    this._configData = await ConfigStore.loadData(this._page.req, this.getGUID() + "_" + super.getValue("ConfigID", 0), "grid");
    return this._configData;
  }

  initPagersDataSource(pageTag, dataSource) {
    var bottomPagerID = this.getID() + "_pager1";
    var topPagerID = this.getID() + "_pager2";
    var pager = this.getValue("Pager", "None");
    if (pager == "TopAndBottom") {
      this.initDataSource(pageTag, topPagerID, dataSource); // 顶部pager初始化（必须同时初始化两个pager）
    } else if (pager == "Top") {
      bottomPagerID = topPagerID;
    }

    return this.initDataSource(pageTag, bottomPagerID, dataSource);
  }

  initDataSource(pageTag, pagerId, dataSource) {
    var pager = new Pager(pageTag);
    var newPageSize = this.getPageSize();
    if (this.getValue("PagerShowAll", false)) {
      newPageSize = -1;
    }
    pager._pageSize = newPageSize;
    pager._id = pagerId;
    pager._dataSource = dataSource;

    pageTag.addGridItemCountDatas(pagerId, pager.getItemCount()); // 数据总行数，翻页控件需要获取（翻页器1）
    var list = pager.getFirstPageItemList(); // 服务端预先返回第一页数据
    pager.dispose();
    return list;
  }

  getPageSize() {
    if (this._configData != null && this._configData.PageSize > 0) {
      return this._configData.PageSize; // 用户自定义每页显示条数
    }

    if (this.getValue("AutoPagerPageSize", false) || this.getValue(
      "AutoDefaultRowCount", false)) {
      var height = this.getValue("Height", 300);
      if (height > 0) // 不包括边距的
      {
        var bodyHeight = height - this.getNonBodyHeight() - 1 - this._cHScrollBarHeight; // 1个像素是第一行的顶部线条
        var bodyRows = parseInt((bodyHeight + 1) / this._defaultRowHeight); // 加1像素说明允许超出1像素
        return bodyRows > 1 ? bodyRows : 2;
      }
    }
    return this.getValue("PagerPageSize", settings.GridPagerDefaultPageSize);
  }

  getNonBodyHeight() {
    return this._cDefaultGridHeaderHeight + // 表头
      (this.getValue("ShowHScrollBar", false) ? this._cHScrollBarHeight : 0) + // 加横向滚动条高度
      (this.getValue("ShowFooter", false) ? this._cFooterHeight : 0);
  }
}

class ListView extends BaseStruct {
  constructor(attr) {
    super(attr);
    this._scriptTypeName = "ListView";
  }

  init(pageTag) {
    super.init(pageTag);
  }
}

class FlowPanel extends BaseStruct {
  constructor(attr) {
    super(attr);
    this._scriptTypeName = "FlowPanel";
  }

  async init(pageTag) {
    super.init(pageTag);
    pageTag.addConfigData(this.getID(), await this.getConfigData());
  }

  async getConfigData() {
    var saveConfig = super.getValue("AllowConfig", false);
    if (!saveConfig) return null;

    this._configData = await ConfigStore.loadData(this._page.req, this.getGUID() + "_" + super.getValue("ConfigID", 0), "panel");
    return this._configData;
  }
}

class Style extends BaseStruct {
  constructor(attr) {
    super(attr);
  }

  init(pageTag) {
    super.init(pageTag);
    pageTag.addStyle(this);
  }
}

class Pager {
  constructor(pageTag) {
    this._pageTag = pageTag;
    this._itemCountCache = -1;
    this._showAll = false;
    this._inited = false;
  }

  dispose() {
    this._pageTag = null;
    this._dataSource = null;
  }

  getItemCount(dataSource) {
    this.initDataSource();
    if (this._itemCountCache < 0) { // 避免在 ShowAll 情况下，在Grid中调用一次，Pager中又调用一次
      this._itemCountCache = this._dataSource.getItemCount();
    }
    return this._itemCountCache;
  }

  getFirstPageItemList() {
    this.initDataSource();
    this._dataSource.invalidateCache(); // 第一次取都是重新刷新
    var newPageSize = this._pageSize;
    if (this._showAll) {
      newPageSize = this.getItemCount();
    }
    var interceptor = this._dataSource && this._dataSource.Interceptor ? this._dataSource.Interceptor : null;
    return this.wrapInterceptor(interceptor, 'getItemList', this._dataSource, [0, newPageSize])
    //return this._dataSource.getItemList(0, newPageSize);
  }

  wrapInterceptor(interceptor, funname, obj, args) {
    return async (cb) => {
      try {
        if (interceptor) {
          await interceptor.before(obj, funname, args);
        }
        let rt = await obj[funname].apply(obj, args);
        if (typeof rt === 'function') {
          await rt(async (result, errormsg) => {
            try {
              if (interceptor) {
                await interceptor.after(obj, funname, args, result, errormsg);
              }
              cb(result, errormsg);
            } catch (e) {
              cb(null, e);
            }
          });
        } else {
          cb(rt);
        }
      } catch (e) {
        cb(null, e);
      }
    }
  }

  initDataSource() {
    if (this._inited) return;
    this._inited = true;

    var page = this._pageTag;
    var httpContext = page.httpContext;
    var req = httpContext.req;
    var session = req.session;
    var dataSource = this._dataSource;
    var clientId = page.getControlClientID(this._id);
    dataSource._pageUrl = req.parsed.path;

    var key = sysConsts.SessionPrefixPagerInfo + clientId;

    var pagerInfoData = req.session[key];
    if (pagerInfoData != null) {
      pagerInfoData._createdCount++;
      pagerInfoData._cacheId = session.id + clientId;
      if (req.headers["server-reqesut"] !== "true") {
        pagerInfoData._oldQueryParams = null;
        pagerInfoData._oldItemCount = 0;
        pagerInfoData._isFirstLoad = true;
      }
      cacheManager.add(pagerInfoData._cacheId, new PagerInfo(session, clientId, dataSource), session.id); // 有时候session还在，但缓存丢失了，没找到重现规律，重新加一次缓存(覆盖的)
    } else {
      pagerInfoData = new PagerInfoData(req.session, clientId, dataSource, req.url);
    }
    req.session[key] = pagerInfoData; // 只能存key-value，不能存function，因此改为：PagerInfoData
  }
}
